# standard library imports
import unittest

# local imports
from probeye.postprocessing.sampling import create_pair_plot
from probeye.postprocessing.sampling import create_posterior_plot
from probeye.postprocessing.sampling import create_trace_plot


class TestProblem(unittest.TestCase):
    def test_not_implemented_pair_plot(self):
        # check for currently not implemented plotting frameworks
        with self.assertRaises(NotImplementedError):
            create_pair_plot(None, None, plot_with="seaborn")
        with self.assertRaises(NotImplementedError):
            create_pair_plot(None, None, plot_with="matplotlib")
        with self.assertRaises(RuntimeError):
            create_pair_plot(None, None, plot_with="other package")

    def test_not_implemented_create_posterior_plot(self):
        # check for currently not implemented plotting frameworks
        with self.assertRaises(NotImplementedError):
            create_posterior_plot(None, None, plot_with="seaborn")
        with self.assertRaises(NotImplementedError):
            create_posterior_plot(None, None, plot_with="matplotlib")
        with self.assertRaises(RuntimeError):
            create_posterior_plot(None, None, plot_with="other package")

    def test_not_implemented_create_trace_plot(self):
        # check for currently not implemented plotting frameworks
        with self.assertRaises(NotImplementedError):
            create_trace_plot(None, None, plot_with="seaborn")
        with self.assertRaises(NotImplementedError):
            create_trace_plot(None, None, plot_with="matplotlib")
        with self.assertRaises(RuntimeError):
            create_trace_plot(None, None, plot_with="other package")


if __name__ == "__main__":
    unittest.main()



# standard library
import logging
import unittest

# third party imports
import numpy as np

# local imports
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor
from probeye.definition.inference_problem import InferenceProblem
from probeye.definition.noise_model import NormalNoiseModel
from probeye.inference.emcee_.solver import EmceeSolver


class TestProblem(unittest.TestCase):
    def test_emcee_solver(self):

        np.random.seed(6174)

        # define the forward model
        class LinRe(ForwardModelBase):
            def __call__(self, inp):
                x = inp["x"]
                a = inp["a"]
                b = inp["b"]
                return {"y": a * x + b}

        # set up the problem
        problem = InferenceProblem("Linear regression")
        problem.add_parameter("a", "model", prior=("normal", {"loc": 0, "scale": 1}))
        problem.add_parameter("b", "model", prior=("normal", {"loc": 0, "scale": 1}))
        problem.add_parameter(
            "sigma", "noise", prior=("uniform", {"low": 0.1, "high": 1})
        )
        problem.add_forward_model(
            "LinRe", LinRe(["a", "b"], [Sensor("x")], [Sensor("y")])
        )
        problem.add_noise_model(NormalNoiseModel({"sigma": "std"}, sensors=Sensor("y")))

        # generate and add some simple test data
        n_tests, a_true, b_true, sigma_true = 5000, 0.3, -0.2, 0.1
        x_test = np.linspace(0.0, 1.0, n_tests)
        y_true = a_true * x_test + b_true
        y_test = np.random.normal(loc=y_true, scale=sigma_true)
        problem.add_experiment(
            f"Tests", fwd_model_name="LinRe", sensor_values={"x": x_test, "y": y_test}
        )

        # run the emcee solver with deactivated output
        logging.root.disabled = True
        emcee_solver = EmceeSolver(problem, show_progress=False, seed=6174)
        _ = emcee_solver.run_mcmc(n_walkers=20, n_steps=200, vectorize=False)
        # summary = run_emcee_postprocessing(problem, emcee_sampler,
        #                                    show_progress=True)
        # sample_means = summary['mean']
        # for mean, mean_true\
        #         in zip(sample_means, [a_true, b_true, sigma_true]):
        #     self.assertAlmostEqual(mean, mean_true, delta=0.01)


if __name__ == "__main__":
    unittest.main()



# standard library
from typing import Optional, TYPE_CHECKING

# third party imports
import matplotlib.pyplot as plt

# local imports (inference engines)
from probeye.inference.scipy_.solver import ScipySolver
from probeye.inference.emcee_.solver import EmceeSolver
from probeye.inference.torch_.solver import PyroSolver

# local imports (post-processing)
from probeye.postprocessing.sampling import create_pair_plot
from probeye.postprocessing.sampling import create_posterior_plot
from probeye.postprocessing.sampling import create_trace_plot

# imports only needed for type hints
if TYPE_CHECKING:
    from probeye.definition.inference_problem import InferenceProblem


def run_inference_engines(
    problem: "InferenceProblem",
    true_values: Optional[dict] = None,
    n_steps: int = 1000,
    n_initial_steps: int = 100,
    n_walkers: int = 20,
    plot: bool = True,
    show_progress: bool = True,
    run_scipy: bool = True,
    run_emcee: bool = True,
    run_torch: bool = True,
):
    """
    Runs a requested selection of inference engines on a given problem. This function is
    in a separate file in order to avoid repeating the same lines of code in all the
    integration tests.

    Parameters
    ----------
    problem
        Describes the inference problem including e.g. parameters and data.
    true_values
        Used for plotting 'true' parameter values. Keys are the parameter names and
        values are the values that are supposed to be shown in the marginal plots.
    n_steps
        Number of steps (samples) to run.
    n_initial_steps
        Number of steps for initial (burn-in) sampling.
    n_walkers
        Number of walkers used by the estimator.
    plot
        If True, the data and the post-processing plots are plotted.
    show_progress
        If True, progress-bars will be shown, if available.
    run_scipy
        If True, the problem is solved with scipy (maximum likelihood estimate).
        Otherwise, no maximum likelihood estimate is derived.
    run_emcee
        If True, the problem is solved with the emcee solver. Otherwise, the
        emcee solver will not be used.
    run_torch
        If True, the problem is solved with the pyro/torch_ solver. Otherwise, the
        pyro/torch_ solver will not be used.
    """

    # this loop avoids to write down the same lines of code for the post-processing for
    # each inference engine again
    for inference_engine, requested_to_run in {
        "scipy": run_scipy,
        "emcee": run_emcee,
        "pyro": run_torch,
    }.items():

        # solve the problem with scipy (max. likelihood) if requested
        if inference_engine == "scipy":
            is_sampling_solver = False
            if requested_to_run:
                scipy_solver = ScipySolver(problem, show_progress=show_progress)
                inference_data = scipy_solver.run_max_likelihood(
                    true_values=true_values
                )
            else:
                # in this case, the engine was not requested to run
                continue

        # solve the problem with emcee if requested
        elif inference_engine == "emcee":
            is_sampling_solver = True
            if requested_to_run:
                emcee_solver = EmceeSolver(problem, show_progress=show_progress)
                inference_data = emcee_solver.run_mcmc(
                    n_walkers=n_walkers,
                    n_steps=n_steps,
                    n_initial_steps=n_initial_steps,
                )
            else:
                # in this case, the engine was not requested to run
                continue

        # solver the problem with pyro/torch if requested
        elif inference_engine == "pyro":
            is_sampling_solver = True
            if requested_to_run:
                n_walkers_used = 1  # getting errors when trying to use more
                pyro_solver = PyroSolver(problem, show_progress=show_progress)
                inference_data = pyro_solver.run_mcmc(
                    n_walkers=n_walkers_used,
                    n_steps=n_steps,
                    n_initial_steps=n_initial_steps,
                )
            else:
                # in this case, the engine was not requested to run
                continue

        else:
            # if this happens, you messed something up in the defining
            # dictionary of this loop
            raise RuntimeError(f"Found unknown inference engine '{inference_engine}'!")

        # do the post-processing; note that the interface is the same for each inference
        # engine; also note, that the plots are not so much intended for automatic
        # testing, as for manually running the script
        if is_sampling_solver:
            create_pair_plot(
                inference_data, problem, true_values=true_values, show=False
            )
            create_pair_plot(
                inference_data,
                problem,
                focus_on_posterior=True,
                kind="hexbin",
                true_values=true_values,
                show=False,
                marginal_kwargs={"kind": "hist", "hist_kwargs": {"bins": 10}},
            )
            create_pair_plot(
                inference_data,
                problem,
                plot_priors=False,
                kind="scatter",
                true_values=true_values,
                show=False,
                marginal_kwargs={"kind": "hist", "hist_kwargs": {"bins": 10}},
            )
            create_pair_plot(
                inference_data,
                problem,
                plot_priors=False,
                true_values=true_values,
                show=False,
            )
            create_posterior_plot(
                inference_data, problem, true_values=true_values, show=False
            )
            create_trace_plot(inference_data, problem, show=False)
            if plot:
                plt.show()  # shows all plots at once due to 'show=False' above
